@@ -2,7 +2,7 @@ import numpy as np
 from math import log2
 from scipy.cluster import hierarchy
 from sklearn.cluster import AgglomerativeClustering
-from sklearn.neighbors import kneighbors_graph
+from sklearn.neighbors import kneighbors_graph, NearestNeighbors
 from sklearn.metrics.pairwise import pairwise_kernels
 from scipy import sparse
 import heapq
@@ -83,12 +83,14 @@ class pySETree():
                  objective='SE',
                  strategy='top_down',
                  split_se_cutoff=0.05,
+                 verbose=False
                  ):
         self.strategy = strategy
         self.objective = objective
         self.min_k = min_k
         self.max_k = max_k
         self.split_se_cutoff = split_se_cutoff
+        self.verbose = verbose
 
         self.vertex_num = aff_m.shape[0]
         if self.max_k > self.vertex_num:
@@ -221,13 +223,13 @@ class pySETree():
                 node1_vs = np.argwhere(np.array(clusters) == 0).T[0]
                 node2_vs = np.argwhere(np.array(clusters) == 1).T[0]
 
-            if len(fiedler_vector) != self.vertex_num:
-                node1_mean = np.mean(self.root_fiedler_vector[node1_vs])
-                node2_mean = np.mean(self.root_fiedler_vector[node2_vs])
-                if node1_mean > node2_mean:
-                    tmp = node1_vs
-                    node1_vs = node2_vs
-                    node2_vs = tmp
+                if len(fiedler_vector) != self.vertex_num:
+                    node1_mean = np.mean(self.root_fiedler_vector[node1_vs])
+                    node2_mean = np.mean(self.root_fiedler_vector[node2_vs])
+                    if node1_mean > node2_mean:
+                        tmp = node1_vs
+                        node1_vs = node2_vs
+                        node2_vs = tmp
 
         else:  # 2 vs
             Z[node.id-(self.vertex_num)] = [node.vs[0], node.vs[1], -1, 2, node.id]
@@ -335,7 +337,8 @@ class pySETree():
                 delta = graph_metric.get_delta_nm(M, sparse_m, vG, root, node1, node2)
             heapq.heappush(heap, (-delta, n1, n2))
             i += 1
-        print('i', i, '(linkage - initial non zero pair)')
+        if self.verbose:
+            print('i', i, '(linkage - initial non zero pair)')
 
         z_i = 0
         count = 0
@@ -345,13 +348,16 @@ class pySETree():
             while individuals:
                 if not self.G.edges(data=True):
                     break
-                print('-----', self.objective, 'merge phase', count, only_positive_delta, 'start, individuals: ', len(individuals), ', leaves ', len(leaves))
-                print(len(self.G.edges(data=True)))
+                if self.verbose:
+                    print('-----', self.objective, 'merge phase', count, only_positive_delta, 'start, individuals: ', len(individuals), ', leaves ', len(leaves))
+                    print(len(self.G.edges(data=True)))
                 z_i, merge_phase_i = self._merge_phase(root, individuals, leaves, Z, z_i, heap,
                                                        only_positive_delta=only_positive_delta)
-                print('-----', self.objective, 'merge phase', count, 'end, individuals', len(individuals), ', leaves ', len(leaves))
+                if self.verbose:
+                    print('-----', self.objective, 'merge phase', count, 'end, individuals', len(individuals), ', leaves ', len(leaves))
                 i += merge_phase_i
-                print('i', i, 'merge phase')
+                if self.verbose:
+                    print('i', i, 'merge phase')
                 count += 1
                 if len(individuals) > 0:
                     break
@@ -359,18 +365,22 @@ class pySETree():
                     self.node_list[l].is_individual = True
                     individuals[l] = 1
                     i += 1
-                print('i', i, 'merge phase update leaves states')
+                if self.verbose:
+                    print('i', i, 'merge phase update leaves states')
 
         self.leaves = [l for l in leaves]
         while z_i < self.vertex_num - 1:
-            print('-----', self.objective, 'binary merge', count)
+            if self.verbose:
+                print('-----', self.objective, 'binary merge', count)
             z_i, binary_combine_i = self._binary_combine(root, leaves, Z, z_i,
                                                          only_positive_delta=False)
             i += binary_combine_i
-            print('i', i, 'binary_combine')
+            if self.verbose:
+                print('i', i, 'binary_combine')
             count += 1
 
-        print('N', N, 'i', i)
+        if self.verbose:
+            print('N', N, 'i', i)
         return Z
 
     def _merge_phase(self, root, individuals, leaves, Z, z_i, heap,
@@ -544,8 +554,9 @@ class pySETree():
         cutoff_m.fill(-1)
 
         self._dp_compute_cost(cost_m, cutoff_m, node_ids)
-        print('se cost')
-        print(cost_m[-1, :])
+        if self.verbose:
+            print('se cost')
+            print(cost_m[-1, :])
         # print(cost_m)
         # print(cost_m.shape)
         # print(cutoff_m)
@@ -570,7 +581,8 @@ class pySETree():
             ks_subpopulation_node_ids.append(subpopulation_node_ids)
 
         optimal_k = self.max_k - np.argmin(cost_m[-1, 2:][::-1])
-        print(cost_m[-1, :][optimal_k])
+        if self.verbose:
+            print(cost_m[-1, :][optimal_k])
         return cost_m[-1, 1:], ks_subpopulation_node_ids, ks_clusters, optimal_k
 
     def _dp_compute_cost(self, cost_m, cutoff_m, node_ids, contract_type='2D'):
@@ -641,7 +653,8 @@ class pySETree():
         return 0
         tree_se = 0
         self.get_tree_se_aux(self.root, tree_se)
-        print(tree_se)
+        if self.verbose:
+            print(tree_se)
 
     def get_tree_se_aux(self, node, tree_se):
         tree_se += node.se
@@ -716,6 +729,10 @@ class SEAT(AgglomerativeClustering):
                  dist_topk=5,
                  split_se_cutoff=0.05,
                  kernel_gamma=None,
+                 outlier_detection=None,
+                 outlier_percentile=None,
+                 outlier_distance=0.5,
+                 verbose=False,
                  ):
         self.min_k = min_k
         self.max_k = max_k
@@ -733,6 +750,11 @@ class SEAT(AgglomerativeClustering):
 
         self.knn_m = knn_m
 
+        self.outlier_detection = outlier_detection
+        self.outlier_percentile = outlier_percentile
+        self.outlier_distance = outlier_distance
+        self.verbose = verbose
+
     def construct_affinity(self, X):
         # https://scikit-learn.org/stable/modules/metrics.html
 
@@ -750,7 +772,8 @@ class SEAT(AgglomerativeClustering):
                 sigma = self.kernel_gamma
             else:
                 sigma = X.std()
-            print('sigma', sigma)
+            if self.verbose:
+                print('sigma', sigma)
             gamma = 1/(sigma*sigma)
             aff_m = pairwise_kernels(X, metric='rbf', gamma=gamma)
 
@@ -833,6 +856,58 @@ class SEAT(AgglomerativeClustering):
 
         self.knn_m = knn_m
 
+    def detect_outlier(self, X):
+        
+        if self.outlier_detection == 'knn_neighbors':
+            knn = NearestNeighbors(n_neighbors=self.n_neighbors)
+            knn.fit(X)
+            distances, indexes = knn.kneighbors(X)
+            distances_mean = distances.mean(axis=1)
+            if self.outlier_percentile:
+                cutoff = np.quantile(distances_mean, self.outlier_percentile)
+            else:
+                cutoff = self.outlier_distance
+            self.outlier_index = np.where(distances_mean > cutoff)[0]
+            self.inlier_index = np.where(distances_mean <= cutoff)[0]
+            return X[self.inlier_index, :]
+        else:
+            self.outlier_index = []
+            return X
+
+    def insert_outliers(self, labels):
+        #print(len(self.outlier_index), len(self.inlier_index))
+        if len(self.outlier_index) == 0:
+            return labels
+        
+        n = len(self.outlier_index) + len(self.inlier_index)
+        new_labels = []
+        inliner_i = 0
+        for i in range(n):
+            if i in self.outlier_index:
+                new_labels.append(-1)
+            else:
+                new_labels.append(labels[inliner_i])
+                inliner_i += 1
+        return new_labels
+
+    def insert_outliers_df(self, label_df):
+        #print(len(self.outlier_index), len(self.inlier_index))
+        if len(self.outlier_index) == 0:
+            return label_df
+        label_m = label_df.values.tolist()
+       
+        n = len(self.outlier_index) + len(self.inlier_index)
+        new_label_m = []
+        inliner_i = 0
+        for i in range(n):
+            if i in self.outlier_index:
+                new_label_m.append([-1]*label_df.shape[1])
+            else:
+                new_label_m.append(label_m[inliner_i])
+                inliner_i += 1
+        new_label_df = pd.DataFrame(new_label_m, columns = label_df.columns)
+        return new_label_df
+        
     def fit(self, X, y=None):
 
         X = self._validate_data(X, ensure_min_samples=2, estimator=self)
@@ -854,13 +929,21 @@ class SEAT(AgglomerativeClustering):
             raise ValueError("sparsification should be affinity, precomputed, knn_neighbors, knn_neighbors_from_X."
                              " %s was provided." % str(self.sparsification))
 
+        if self.outlier_detection not in [None, 'knn_neighbors']:
+            raise ValueError("outlier_detection should be None, precomputed."
+                             " %s was provided." % str(self.outlier_detection))            
+            
         if self.strategy not in ['bottom_up', 'top_down']:
             raise ValueError("affinity should be bottom_up, top_down."
                              " %s was provided." % str(self.strategy))
-
-        print('fit', self.strategy)
+            
+        if self.verbose:
+            print('fit', self.strategy)
+            
+        X = self.detect_outlier(X)
         self.construct_affinity(X)
         self.graph_sparsification(X)
+        
 
         # build the tree
 
@@ -871,12 +954,14 @@ class SEAT(AgglomerativeClustering):
                                self.min_k, self.max_k,
                                self.objective,
                                self.strategy,
-                               self.split_se_cutoff)
+                               self.split_se_cutoff,
+                               verbose=self.verbose)
         self.se_tree = se_tree
         t1 = time.time()
         Z = se_tree.build_tree()
         t2 = time.time()
-        print('build tree time', t2 - t1)
+        if self.verbose:
+            print('build tree time', t2 - t1)
         self.aff_m = se_tree.aff_m
 
         # se_tree.order_tree()
@@ -888,13 +973,14 @@ class SEAT(AgglomerativeClustering):
         self.se_scores = se_tree.se_scores
         self.delta_se_scores = se_tree.delta_se_scores
         self.optimal_k = se_tree.optimal_k
-        self.labels_ = se_tree.optimal_clusters
+        
+        self.labels_ = self.insert_outliers(se_tree.optimal_clusters)
         self.Z_ = Z[:, :4]
         self.leaves_list = hierarchy.leaves_list(self.Z_)
-        self.order = self._order()
-        self.ks_clusters = se_tree.ks_clusters
-        self.Z_clusters = se_tree.Z_clusters
-        self.clubs = self._get_clubs()
+        self.order = self.insert_outliers(self._order())
+        self.ks_clusters = self.insert_outliers_df(se_tree.ks_clusters)
+        self.Z_clusters = self.insert_outliers_df(se_tree.Z_clusters)
+        self.clubs = self.insert_outliers(self._get_clubs())
         self.club_k = len(se_tree.leaves)
 
         self.newick = se_tree.to_newick()
